module Source = struct
  let file = "file"
  let command_line = "command-line"
  let path = "path"
  let ocamlfind_query = "ocamlfind query"
  let ocamldep = "ocamldep"
  let target_pattern = "target pattern"
  let builtin = "builtin configuration"
  let configuration = "configuration"
  let plugin_tag = "plugin tag"
end
